#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Python module to transparently open compressed files. """

import os, io

class TransparentOpen(object):
    """ File opener class with transparent support for compressed files.
        Supported compression types: .gz, .bz2, .lzma, .xz, .7z, .zip,
        .tar, .tar.gz, .tgz, .tar.bz2, .tbz, .tbz2, .tar.xz, .txz,
        .tar.lzma, .tlz.
    """
    def __init__(self, filename, file_in_archive=0, password=None):
        """ Usage: t = TransparentOpen('filename', file_in_archive='abc.txt', password='xxx')

            Returns an object t with a filehandle t.fh which points to the
            decompressed file. In the case of a multifile archive (.zip,
            .7z, and any of the tar combinations), a t.fh_archive will point to
            the opened outer archive, while t.fh points to the requested file
            inside the archive. The filename is stored in t.filename.
            
            file_in_archive can be used to select a file from a multifile
            archive, either by index (0 is the first file in the archive),
            or by filename. On python 2, the filename must be a byte string.
            
            Optionally set the password for encrypted archives. Passwords are
            only supported on zip and 7zip archives.
            
            Use t.close() to close both inner and outer filehandles at once.
            
            TransparentOpen supports the with statement.
        """
        if not isinstance(file_in_archive, (int, str)):
            raise IOError('file_in_archive must be int or str.')

        self.filename = ''
        self.fh_archive = None

        if isinstance(filename, str):
            base, ext = os.path.splitext(filename)
            base_maybe, ext_maybe = os.path.splitext(base)
            if ext_maybe == '.tar':
                base = base_maybe
                ext = ext_maybe + ext

            self.filename = base
            if ext == '.gz':
                import gzip
                self.fh = gzip.open(filename, mode='rb')

            elif ext == '.bz2':
                import bz2
                self.fh = bz2.BZ2File(filename, mode='rb')

            elif ext in ('.lzma', '.xz'):
                # available in Python 3.3 and higher
                try:
                    import lzma
                except ImportError:
                    msg = 'LZMA module is not installed on your system. Use Python 3.3 or higher,'
                    msg += ' or install a module to handle lzma/xz compressed files.'
                    raise IOError(msg)

                self.fh = lzma.LZMAFile(filename, mode='rb')

            elif ext in ('.tar', '.tar.bz2', '.tbz', '.tbz2', '.tar.gz', '.tgz',
                         '.tar.xz', '.txz', '.tar.lzma', '.tlz'):
                import tarfile

                self.fh_archive = tarfile.open(filename, mode='r')
                if isinstance(file_in_archive, int):
                    if file_in_archive == 0:
                        # No need to scan trough entire tar first
                        m = self.fh_archive.firstmember
                    else:
                        m = self.fh_archive.getmembers()
                        m = m[file_in_archive]
                    self.fh = self.fh_archive.extractfile(m)
                    self.filename = m.name
                else:
                    self.fh = self.fh_archive.extractfile(file_in_archive)
                    self.filename = file_in_archive

            elif ext == '.7z':
                # py7zlib is needed for 7z
                try:
                    import py7zlib
                except ImportError:
                    msg = 'pylzma/py7zlib module is not installed on your system. See'
                    msg += ' http://www.joachim-bauch.de/projects/pylzma'
                    raise IOError(msg)

                with open(filename, mode='rb') as fh_raw:
                    archive = py7zlib.Archive7z(fh_raw, password=password)

                    if isinstance(file_in_archive, int):
                        f = archive.files[file_in_archive]
                    else:
                        f = archive.files_map[file_in_archive]

                    self.fh = io.BytesIO(f.read())

            elif ext == '.zip':
                import zipfile

                fh_archive = zipfile.ZipFile(filename, mode='r')

                if password:
                    fh_archive.setpassword(password)

                if isinstance(file_in_archive, int):
                    infolist = fh_archive.infolist()
                    fh_file = fh_archive.open(infolist[file_in_archive])
                else:
                    fh_file = fh_archive.open(file_in_archive)

                self.fh = io.BytesIO(fh_file.read())
                fh_file.close()
                fh_archive.close()

            else:
                # No compression or unrecognized
                self.fh = open(filename, mode='rb')
                self.filename = filename

        elif hasattr(filename, 'read'):
            if (hasattr(filename, 'seek') and hasattr(filename, 'tell')):
                # Is it in binary mode?
                if 'b' in filename.mode:
                    self.fh = filename
                    self.filename = filename.name
                else:
                    msg = 'Fileobject {} opened in text-mode, reopen with mode="rb".'
                    raise IOError(msg.format(filename))
            else:
                # Read but no seek and/or tell: wrap in BytesIO; let's hope it has a name.
                self.fh = io.BytesIO(filename.read())
                self.filename = filename.name
        else:
            msg = 'Cannot open file {}, don\'t know what it is.'
            raise IOError(msg.format(filename))

    def close(self):
        """ Close the file. """
        self.fh.close()
        if self.fh_archive:
            self.fh_archive.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Python module to read Cameca (nano)SIMS data files. """
from __future__ import print_function, division, absolute_import

import io
import sys
import os
import re
import datetime
import warnings
import collections
import copy
import gzip
import bz2
import zipfile
import tarfile
from struct import unpack
import numpy as np

# Return data in pandas 4D frame, if available, numpy array otherwise
try:
    import pandas as pd
except ImportError:
    pd = None

# available in Python 3.3 and higher
try:
    import lzma
except ImportError:
    lzma = None

# py7zlib is needed for 7z, cannot handle xz/lzma archives compressed by xz
try:
    import py7zlib
except ImportError:
    py7zlib = None

# Local import py2 or 3
# try:
#     from .info import *
#     from .utils import format_species
#     from .transparent import TransparentOpen
# except SystemError:
#     from info import *
#     from utils import format_species
#     from transparent import TransparentOpen
#
# if sys.version_info.major >= 3:
#     unicode = str

from sims.utils import format_species
from sims.transparent import TransparentOpen

__all__ = ['SIMSReader', 'SIMS']

_stage_scan_types = {
    0: 'stage scan',
    1: 'beam scan',
    2: 'image scan',
}

# large scale, movement control, small scale
# spot/line/image, beam/stage, raster/static
_file_types = {
    21: 'depth profile',
    22: 'line scan, stage control',
    26: 'isotope image',
    27: 'image',
    29: 'grain mode image',
    31: 'SIBC file',
    35: 'beam stability file',
    39: 'line scan image',
    40: 'line scan, beam control',
    41: 'stage scan image',
}

_supported_file_types = [ 21, 22, 26, 27, 29, 35, 39, 40, 41 ]

_peakcenter_sides = {
    0: 'left',
    1: 'right',
    2: 'both'
}

_exit_slit_labels = {
    0: 'slit 1',
    1: 'slit 2',
    2: 'slit 3'
}

_exit_slit_size_labels = {
    0: 'normal',
    1: 'large',
    2: 'extra large'
}

_detectors = {
    0: 'EM',
    1: 'FC'
}

class SIMSReader(object):
    """ Base class for reading a SIMS file. """
    def __init__(self, fileobject, filename=''):
        """ This class object does not open or close files, and nothing will
            be read by default. An empty header and data array are created.
            Provides all methods needed for reading a SIMS file.
        """
        self.fh = fileobject
        self.filename = filename
        self.header = {}
        self.data = None
        self._data_corr = None
        self._bo = ''

    def __deepcopy__(self, memo):
        """ Assist deep-copying of this class.
            Everything except the filehandle is copied.
        """
        new = type(self)(None)
        for k in self.__dict__.keys():
            if k in ('fh', 'fh_archive'):
                new.__dict__[k] = None
            else:
                new.__dict__[k] = copy.deepcopy(self.__dict__[k], memo)
        return new

    def peek(self):
        """ Peek into image file and determine basic file information.

            Usage: s.peek()

            Reads first 12 bytes of file opened as s.fh, determines byte order
            (endianess), file type, file version, and header size. Information
            is stored in s.header.
        """
        self.fh.seek(0)
        snip = self.fh.read(12)
        # 256 is an arbitrarily chosen limit, file types are in the 10s.
        if unpack('<i', snip[4:8])[0] < 256:
            self.header['byte order'] = '<'
            self._bo = '<'
        elif unpack('>i', snip[4:8])[0] < 256:
            self.header['byte order'] = '>'
            self._bo = '>'
        else:
            raise TypeError("Cannot determine file endianess.")

        self.header['file version'], self.header['file type'], \
            self.header['header size'] = \
            unpack(self._bo + '3i', snip)

        if self.header['file type'] not in _supported_file_types:
            msg = "File of type {} is not supported at the moment."
            msg = msg.format(self.header['file type'])
            raise NotImplementedError(msg)

    def read_header(self):
        """ Read the image header.

            Usage: s.read_header()

            Reads the header from the file object stored in s.fh. Extracts as
            much information as possible and stores it in a Python dictionary
            in s.header. At least byte order, header size, file type, and file
            version need to be known before the header can be read: use
            s.peek() before s.read_header().
        """
        # Read entire header into memory in one read to minimize Disk I/O.
        self.fh.seek(0)
        hdr = self.fh.read(self.header['header size'])

        # Find several markers in the byte-string
        # Each of these may occur more than once, find last.
        polylist_pos = hdr.rfind(b'Poly_list\x00')
        champslist_pos = hdr.rfind(b'Champs_list\x00')
        offsetlist_pos = hdr.rfind(b'Offset_list\x00')

        # Find first occurance for these.
        analparam_pos = hdr.find(b'Anal_param\x00')
        analparamnano_pos = hdr.find(b'Anal_param_nano\x00')
        analparamnanobis_pos = hdr.find(b'Anal_param_nano_bis\x00')

        # Turn byte-string into BytesIO file-like object; reading and
        # keeping track of where we are is easier that way than trying to
        # slice byte-string as an array and keeping track of indices.
        hdr = io.BytesIO(hdr)

        # Main header
        hdr.seek(12)
        self.header.update(self._main_header(hdr))

        # NanoSIMS header, starts with PolyList/ChampsList/OffsetList
        # The following configurations have been found in the wild, so far:
        # 1. NS header
        # 2. PL, NS header
        # 3. PL, CL, OL, NS header
        # 4. PL, CL, OL, partial NS header, PL, NS header, PL, CL, OL,
        #    partial NS header, PL, NS header
        # Note: I have not seen any *lists with contents (only length 0).
        # From OpenMIMS documentation I know that PolyList is as list of
        # Species dicts, but don't know how to read ChampsList or OffsetList.
        if polylist_pos < 0:
            if self.header['analysis type'].endswith('RTI') \
            or self.header['file type'] == 35:
                # Case 1: No PL marker, so far only found for Real Time Images
                # or beam stability file
                hdr.seek(216, 1)
            else:
                msg = 'No PolyList marker found in header and not and RTI image. '
                msg += 'Don\'t know how to continue.'
                raise NotImplementedError(msg)
        elif (champslist_pos < 0 and offsetlist_pos < 0):
            # Case 2: PL, NS header
            self.header['PolyList'] = self._pco_list(hdr, 'poly', polylist_pos)
        elif (polylist_pos < champslist_pos < offsetlist_pos):
            # Case 3: PL, CL, OL, NS header
            self.header['PolyList'] = self._pco_list(hdr, 'poly', polylist_pos)
            self.header['ChampsList'] = self._pco_list(hdr, 'champs', champslist_pos)
            self.header['OffsetList'] = self._pco_list(hdr, 'offset', offsetlist_pos)
        elif (champslist_pos < offsetlist_pos < polylist_pos):
            # Case 4: PL, CL, OL, partial NS header, PL, NS header
            # with possible repeat
            self.header['ChampsList'] = self._pco_list(hdr, 'champs', champslist_pos)
            self.header['OffsetList'] = self._pco_list(hdr, 'offset', offsetlist_pos)
            self.header['PolyList'] = self._pco_list(hdr, 'poly', polylist_pos)
        else:
            msg = 'An unknown order of the Poly/Champs/Offset Lists occured.\n'
            msg += 'Positions: PL = {}, CL = {}, OL = {}'
            msg = msg.format(polylist_pos, champslist_pos, offsetlist_pos)
            raise NotImplementedError(msg)

        self.header['NanoSIMSHeader'] = self._nanosims_header(hdr)

        # How much to skip? Chomping does not work; what if first value is 0?
        # This is correct so far, for nsheader v8 and 9
        hdr.seek(948, 1)
        self.header['BFields'] = []
        for b in range(self.header['NanoSIMSHeader']['b fields']):
            bf = self._bfield(hdr)
            bf['counting frame time'] = bf['time per pixel'] * \
                self.header['NanoSIMSHeader']['counting frame height'] * \
                self.header['NanoSIMSHeader']['counting frame width']
            bf['scanning frame time'] = bf['time per pixel'] * \
                self.header['NanoSIMSHeader']['scanning frame height'] * \
                self.header['NanoSIMSHeader']['scanning frame width']
            bf['working frame time'] = bf['time per pixel'] * \
                self.header['NanoSIMSHeader']['working frame height'] * \
                self.header['NanoSIMSHeader']['working frame width']
            self.header['BFields'].append(bf)
        # End nanosims_header/bfield based on Poly_list position

        # Analytical parameters

        ## anal_param is not in OpenMIMS at all, represents file
        ## Cameca NanoSIMS Data/raw_spec/cur_anal_par
        ## However, only few useful things in this section, all of
        ## which are also in other sections. Skip.
        # if analparam_pos < 0:
        #     msg = 'Anal_param not found in header, skipping.'
        #     warnings.warn(msg)
        # else:
        #     hdr.seek(analparam_pos + 24)
        #     print(analparam_pos)
        #     d = {}
        #     d['primary ion'], d['primary current begin'], \
        #         d['primary current end'], d['raster'], \
        #         d['X 00 always 1.0'], \
        #         d['X 01 always 1'], d['X 02 always 0'], \
        #         d['X 03 always 1'], d['X 04 always 0'], \
        #         d['X 05 always 0'], d['X 06 (not0 always 0'], \
        #         d['X 07 (not) always 0'], d['X 08 always 0'], \
        #         d['pressure 1'], d['e0w'], d['X 09 always 35 or #'], \
        #         d['X 10 junk'], \
        #         d['X 11 always 1'], d['X 12 always 0'], \
        #         d['X 13 always 1'], d['X 14 always 0'], \
        #         d['X 15 always 0'], d['X 16 always 0'], \
        #         d['X 17 always 0'], d['X 18 always 0'], \
        #         d['X 19 always 0'], d['X 20 always 300'], \
        #         d['X 21'], d['X 22'], d['X 23'], d['X 24'], \
        #     d['pressure 2'], d['X 25 junk'] = \
        #     unpack(self._bo + '24s 4d 8i 48s d i 28s 14i 8s 176s', hdr.read(416))
        #
        #     d['pressure 1'] = self._cleanup_string(d['pressure 1'])
        #     d['pressure 2'] = self._cleanup_string(d['pressure 2'])
        #     d['primary ion'] = self._cleanup_string(d['primary ion'])
        #
        #     self.header['AnalParam'] = d

        # Called AnalyticalParamNano AND AnalysisParamNano in OpenMIMS.
        # Here, split out Primary and Secondary beam.
        # Represents the file Cameca NanoSIMS Data/raw_spec/cur_anal_par_nano
        if analparamnano_pos < 0:
            msg = 'Anal_param_nano not found in header, '
            msg += 'don\'t know where PrimaryBeam section starts.'
            warnings.warn(msg)
        else:
            hdr.seek(analparamnano_pos + 16)
            self.header['analysis version'], self.header['n50large'], \
                self.header['comment'] = \
                unpack(self._bo + '2i 8x 256s', hdr.read(272))

            self.header['n50large'] = bool(self.header['n50large'])
            self.header['comment'] = self._cleanup_string(self.header['comment'])

            self.header['PrimaryBeam'] = self._primary_beam(hdr)
            self.header['SecondaryBeam'] = self._secondary_beam(hdr)
            self.header['Detectors'] = self._detectors1(hdr)

            self.header['SecondaryBeam']['E0S'] = self.header['Detectors'].pop('E0S')
            self.header['SecondaryBeam']['pressure multicollection chamber'] = \
                self.header['Detectors'].pop('pressure multicollection chamber')

            # Add overall mode of machine, based on E0W
            if self.header['SecondaryBeam']['E0W'] < 0:
                self.header['polarity'] = '+'
            else:
                self.header['polarity'] = '-'

            # Combine pixel size from NanoSIMSHeader and raster from PrimaryBeam
            self.header['NanoSIMSHeader']['working frame raster'] = \
                self.header['PrimaryBeam']['raster']
            self.header['NanoSIMSHeader']['scanning frame raster'] = \
                self.header['NanoSIMSHeader']['working frame raster'] * \
                self.header['NanoSIMSHeader']['scanning frame width'] / \
                self.header['NanoSIMSHeader']['working frame width']
            self.header['NanoSIMSHeader']['counting frame raster'] = \
                self.header['NanoSIMSHeader']['working frame raster'] * \
                self.header['NanoSIMSHeader']['counting frame width'] / \
                self.header['NanoSIMSHeader']['working frame width']

            # Header for non-nano SIMS
            magic = unpack(self._bo + 'i', hdr.read(4))[0]
            if magic != 2306:
                msg = 'SIMSHeader magic number not found here at byte {}.'
                msg = msg.format(hdr.tell()-4)
                raise ValueError(msg)
            self.header['SIMSHeader'] = self._sims_header(hdr)

            if self.header['analysis version'] >= 5:
                if analparamnanobis_pos < 0:
                    msg = 'Anal_param_nano_bis not found in header, '
                    msg += 'don\'t know where second Detectors section starts.'
                    warnings.warn(msg)
                else:
                    hdr.seek(analparamnanobis_pos + 24)
                    self.header['Detectors'].update(self._detectors2(hdr))
                    xl = self.header['Detectors'].pop('exit slit xl')
                    for n in range(7):
                        det = self.header['Detectors']['Detector {}'.format(n+1)]
                        w = list(det['exit slit widths'])
                        w[2] = xl[5*n:5*(n+1)]
                        det['exit slit widths'] = tuple(w)
                        h = list(det['exit slit heights'])
                        h[2] = xl[5*(n+1):5*(n+2)]
                        det['exit slit heights'] = tuple(h)

                # Presets
                self.header['Presets'] = self._presets(hdr)

            # End Detectors pt 2 based on anal_param_nano_bis position

            # Last part of detectors
            if self.header['analysis version'] >= 6:
                d3 = self._detectors3(hdr)
                self.header['Detectors']['TIC'] = d3.pop('TIC')
                for k, v in d3.items():
                    self.header['Detectors'][k].update(v)
        # End PrimaryBeam/SecondaryBeam/Presets/Detectors based on anal_param_nano position

        # Image header, at end of overall header
        if self.header['file type'] == 26:
            hdr.seek(-176, 2)
            self.header['Isotopes'] = self._isotopes_hdr(hdr)
        elif self.header['file type'] in (21, 22, 35):
            # no image header for line scan or beam stability
            pass
        else:
            hdr.seek(-84, 2)
            self.header['Image'] = self._image_hdr(hdr)

        # Done reading header. Check for and read external files for extra info.
        # if os.path.exists(os.path.splitext(self.filename)[0] + '.chk_is'):
        # self._read_chk_is()

    def read_data(self):
        """ Read the image data.

            Usage: s.read_data()

            Reads all the image data from the file object in s.fh. The data is
            stored in s.data. The image header must be read before data can be read.

            Image data are stored as a pandas Panel4D object with species on the
            'labels' axis, frames on the 'items' axis, rows (stage X-axis) on the
            'major' axis, and columns (stage Y-axis) on the 'minor' axis. If pandas
            is not avaible, a Numpy array with 4 dimensions is stored: species, frame,
            row, column.

            Isotope data are stored as a pandas DataFrame with frame number as index
            and species labels as column headers. If pandas is not available, a numpy
            array is stored. Data read from the .is file were corrected and converted
            by the Cameca software during the measurement. These data are stored as
            s._data_corr. If a .is_txt file is present, then the raw data are read
            from the .is_txt file and stored as s.data in counts/s.
        """
        if not self.header['data included']:
            pass
        elif self.header['file type'] in (21, 26):
            self._isotope_data()
            if os.path.exists(self.filename + '_txt'):
                self._isotope_txt_data()
        elif self.header['file type'] == 22:
            # line scan types, no ImageHeader
            warnings.warn('No data read for line scan, fix')
            pass
        elif self.header['file type'] == 35:
            self._beamstability_data()
        else:
            self._image_data()

    def copy(self):
        """ Return a copy of the SIMS object.

            Usage: s2 = s.copy()

            The copy is always a deepcopy, meaning that all data and the full
            header are copied, there are no references to the original. This
            way the data of the copy can be altered without the data of the
            original being changed as well.

            The only exception is the filehandle(s) fh (and fh_archive): they
            cannot be copied. File operations such as read_header and
            read_data is therefore not possible after copy.
        """
        return copy.deepcopy(self)

    def _main_header(self, hdr):
        """ Internal function; reads variable number of bytes; returns main header dict """
        d = {}
        # Called readDefAnalysis in OpenMIMS
        d['sample type'], d['data included'], d['sample x'], d['sample y'], \
            d['analysis type'], d['user name'], d['sample z'], date, time = \
            unpack(self._bo + '4i 32s 16s i 12x 16s 16s', hdr.read(112))

        d['data included'] = bool(d['data included'])
        d['user name'] = self._cleanup_string(d['user name'])
        d['analysis type'] = self._cleanup_string(d['analysis type']).lower()
        date = self._cleanup_string(date)
        time = self._cleanup_string(time)
        d['date'] = self._cleanup_date(date + ' ' + time)

        if self.header['file type'] in (27, 29, 39):
            # Called MaskImage/readMaskIm in OpenMIMS
            d['original filename'], d['analysis duration'], d['frames'], d['scan type'], \
                d['magnification'], d['size type'], d['size detector'], \
                d['beam blanking'], d['presputtering'], d['presputtering duration'] = \
                unpack(self._bo + '16s 3i 3h 2x 3i', hdr.read(48))

            d['AutoCal'] = self._autocal(hdr)
            d['HVControl'] = {}
            d['HVControl']['hvcontrol enabled'] = False

        elif self.header['file type'] in (22, 41):
            # Called MaskSampleStageImage/readMaskIss in OpenMIMS
            d['original filename'], d['analysis duration'], d['scan type'], \
                d['steps'], d['step size x'], d['step size y'], d['step size?'], \
                d['step waittime'], d['frames'], d['beam blanking'], \
                d['presputtering'], d['presputtering duration'] = \
                unpack(self._bo + '16s 6i d 4i', hdr.read(64))

            d['scan type'] = _stage_scan_types.get(d['scan type'], str(d['scan type']))

            d['AutoCal'] = self._autocal(hdr)
            d['HVControl'] = self._hvcontrol(hdr)
            # Don't know if this unused byte needs to go after HVControl or after SigRef.
            hdr.seek(4, 1)

        elif self.header['file type'] in (21, 26):
            # Not in OpenMIMS
            # this bit same as image, 1 extra unused/unknown
            d['original filename'], d['analysis duration'], d['frames'], d['scan type'], \
                d['magnification'], d['size type'], d['size detector'], \
                d['beam blanking'], d['presputtering'], d['presputtering duration'] = \
                unpack(self._bo + '16s 4x 3i 3h 2x 3i', hdr.read(52))

            # this bit same as stage scan
            d['AutoCal'] = self._autocal(hdr)
            d['HVControl'] = self._hvcontrol(hdr)

            # 24 bytes unknown, not sure if they go here or before AutoCal
            hdr.seek(24, 1)
        elif self.header['file type'] == 35:
            d['original filename'], unknown, d['analysis duration'], d['frames'] = \
                unpack(self._bo + '16s 3i 48x', hdr.read(76))
            
            # Set these here so that they can be converted later,
            # or until I figure out what else is in 48 skipped bytes.
            d['scan type'] = 0
            d['beam blanking'] = 0
            d['presputtering'] = 0

            d['AutoCal'] = self._autocal(hdr)
            d['HVControl'] = self._hvcontrol(hdr)
        else:
            raise TypeError('What type of image are you? {}'.format(self.header['file type']))

        # Continue main header for all types
        d['SigRef'] = self._sigref(hdr)
        d['masses'] = unpack(self._bo + 'i', hdr.read(4))[0]

        # scan type is set for stage scan analysis, set others
        if isinstance(d['scan type'], int):
            if d['scan type'] == 0:
                d['scan type'] = ''
            else:
                d['scan type'] = str(d['scan type'])

        d['beam blanking'] = bool(d['beam blanking'])
        d['presputtering'] = bool(d['presputtering'])
        d['original filename'] = self._cleanup_string(d['original filename'])

        if self.header['file type'] in (21, 26, 27, 29, 35, 39):
            if self.header['file version'] >= 4108:
                n = 60
            else:
                n = 10
        elif self.header['file type'] in (22, 40, 41):
            n = 20
        else:
            n = 0

        # Not sure what this is, memory pointers? Not needed.
        # d['mass table ptr'] = unpack(self._bo + n*'i', hdr.read(n*4))
        hdr.seek(n*4, 1)

        if self.header['file type'] in (21, 22, 26, 40, 41, 35):
            hdr.seek(4, 1)  # 4 bytes unused

        # Mass table, dict by species label.
        d['MassTable'] = collections.OrderedDict()
        for m in range(d['masses']):
            mi = {}
            mi['trolley index'], unknown, mi['mass'], \
                mi['matrix or trace'], mi['detector'], mi['wait time'], \
                mi['frame count time'], mi['offset'], mi['b field index'] = \
                unpack(self._bo + '2i d 2i 2d 2i', hdr.read(48))

            mi.update(self._species(hdr))

            # Add correction controls, my own addition.
            mi['background corrected'] = False
            mi['deadtime corrected'] = False
            mi['yield corrected'] = False

            label = mi.pop('label')
            # This is true for NS50L and file version 4108. Anywhere else different?
            # Maybe confirm this with the Trolleys dict, there is an Esi trolley.
            if mi['trolley index'] == 8:
                label = 'SE'

            d['MassTable'][label] = mi

        # Create a few convenient lists
        d['label list'] = tuple(d['MassTable'].keys())
        d['label list fmt'] = tuple(format_species(m) for m in d['label list'])
        d['mass list'] = tuple(d['MassTable'][m]['mass'] for m in d['label list'])

        return d

    def _autocal(self, hdr):
        """ Internal function; reads 76 bytes; returns AutoCal dict """
        # Called AutoCal in OpenMIMS source
        # OpenMIMS says extra unused byte after autocal enabled
        # for stage scan image; not true
        d = {}
        d['autocal enabled'], d['label'], d['begin'], d['duration'] = \
            unpack(self._bo + 'i 64s 2i', hdr.read(76))

        d['autocal enabled'] = bool(d['autocal enabled'])
        d['label'] = self._cleanup_string(d['label'])
        return d

    def _hvcontrol(self, hdr):
        """ Internal function; reads 112 bytes, returns HVControl dict. """
        # Called readHvControl by OpenMIMS
        d = {}
        d['hvcontrol enabled'], d['label'], d['begin'], d['duration'], d['limit low'], \
            d['limit high'], d['step'], d['bandpass width'], d['count time'] = \
            unpack(self._bo + 'i 64s 2i 3d i d', hdr.read(112))

        d['hvcontrol enabled'] = bool(d['hvcontrol enabled'])
        d['label'] = self._cleanup_string(d['label'])
        return d

    def _sigref(self, hdr):
        """ Internal function; reads 160 bytes; returns SigRef dict """
        # Called SigRef in OpenMIMS
        d = {}
        d['sigref enabled'] = bool(unpack(self._bo + 'i', hdr.read(4))[0])
        d['Species'] = self._species(hdr)
        d['detector'], d['offset'], d['quantity'] = \
            unpack(self._bo + '3i', hdr.read(12))
        return d

    def _species(self, hdr):
        """ Internal function; reads 144 bytes, return Species dict. """
        # Called PolyAtomic in OpenMIMS source
        d = {}

        # Charge is read in OpenMIMS as a char, which is not a Java char (2-bytes),
        # but a single byte char.
        d['numeric flag'], d['numeric value'], d['elements'], \
            d['charges'], d['charge label'], d['label'] = \
            unpack(self._bo + '4i c 64s', hdr.read(81))

        d['label'] = self._cleanup_string(d['label'])
        d['charge label'] = self._cleanup_string(d['charge label'])

        # OpenMIMS says 3 bytes AFTER el.table are unused; this is wrong,
        # 3 bytes BEFORE el.table (b 81-84) are unused. n_elements (here: atomic number)
        # is element number in periodic table rather than number of elements.
        # n_isotopes (here: isotope number) is offset from main atomic Z number.
        # Also: collapse ElementTable (Tabelts) into main dict, too many layers.
        hdr.seek(3, 1)
        atoms = unpack(self._bo + '15i', hdr.read(60))
        d['atomic number'] = tuple(n for n in atoms[::3])
        d['isotope number'] = tuple(n for n in atoms[1::3])
        d['stoich number'] = tuple(n for n in atoms[2::3])
        return d

    def _pco_list(self, hdr, name, pos):
        """ Internal function; reads 'name'list, returns 'Name'List list.
            Name is one of "poly", "champs", or "offset". pos is the byte-
            position where the maker 'Name'List starts.
        """
        if name not in ('poly', 'champs', 'offset'):
            raise TypeError('Name must be one of "poly", "champs", or "offset".')

        hdr.seek(pos + 16)
        length = unpack(self._bo + 'i', hdr.read(4))[0]
        d = []
        for p in range(length):
            if name == 'poly':
                d.append(self._species(hdr))
            else:
                msg = '{}List is non-null, don\'t know how to read.'.format(name.capitalize())
                raise NotImplementedError(msg)
        hdr.seek(4, 1)
        return d

    def _nanosims_header(self, hdr):
        """ Internal function; reads 604 bytes; returns NanoSIMSHeader dict """
        # Called MaskNano in OpenMIMS; BFieldTab separated out; create extra sub-dict PeakCenter
        d = {}
        d['PeakCenter'] = {}
        d['nanosimsheader version'], d['regulation mode'], d['mode'], \
            d['grain mode'], d['semigraphic mode'], d['stage delta x'], \
            d['stage delta y'], d['working frame width'], d['working frame height'], \
            d['scanning frame x'], d['scanning frame width'], \
            d['scanning frame y'], d['scanning frame height'], \
            d['counting frame x start'], d['counting frame x end'], \
            d['counting frame y start'], d['counting frame y end'], \
            d['detector type'], d['electron scan'], d['scanning mode'], \
            d['beam blanking'], d['PeakCenter']['peakcenter enabled'], \
            d['PeakCenter']['start'], d['PeakCenter']['frequency'], d['b fields'] = \
            unpack(self._bo + '25i', hdr.read(100))

        d['PeakCenter']['peakcenter enabled'] = bool(d['PeakCenter']['peakcenter enabled'])
        d['regulation mode'] = bool(d['regulation mode'])
        d['grain mode'] = bool(d['grain mode'])
        d['semigraphic mode'] = bool(d['semigraphic mode'])
        d['scanning mode'] = bool(d['scanning mode'])

        # Set a few extra variables.
        d['counting frame width'] = d['counting frame x end'] - d['counting frame x start'] + 1
        d['counting frame height'] = d['counting frame y end'] - d['counting frame y start'] + 1

        # Found in at least one version (file v11, nsHeader v8) a repeat of Poly_list and this
        # first part of nanoSIMSHeader. Total of repeat adds up to 288.
        # After last Poly_list, 288 byte padding zone, not all null-bytes.
        hdr.seek(288, 1)

        # Is this the nPrintRed from OpenMIMS?
        d['print results'] = bool(unpack(self._bo + 'i', hdr.read(4))[0])

        d['SibCenterHor'] = self._sib_center(hdr)
        d['SibCenterVert'] = self._sib_center(hdr)

        # Duplicate and store these two in sub dicts
        b_field_index, has_sib_center = \
            unpack(self._bo + '2i', hdr.read(8))
        if b_field_index < 0:
            b_field_index = None
        has_sib_center = bool(has_sib_center)

        d['SibCenterHor']['b field index'] = b_field_index
        d['SibCenterVert']['b field index'] = b_field_index
        d['SibCenterHor']['sib center enabled'] = has_sib_center
        d['SibCenterVert']['sib center enabled'] = has_sib_center

        d['EnergyCenter'] = self._energy_center(hdr)
        d['E0SCenter'] = self._e0s_center(hdr)

        d['EnergyCenter']['wait time'], d['presputtering raster'], \
            d['PeakCenter']['E0P offset'], d['E0SCenter']['steps'], \
            d['baseline measurement'], d['baseline offset'], \
            d['baseline frequency'] = \
            unpack(self._bo + '5i d i', hdr.read(32))
        return d

    def _sib_center(self, hdr):
        """ Internal function; reads 40 bytes; returns SibCenter dict """
        # Called SecIonBeamNano in OpenMIMS
        d = {}
        d['detector'], d['start'], d['step size'], d['center'], \
            d['50% width'], d['count time'] = \
            unpack(self._bo + '3i 4x 2d i 4x', hdr.read(40))

        if d['detector'] < 0:
            d['detector'] = None
        d['count time'] /= 100  # 10 ms increments to seconds
        return d

    def _energy_center(self, hdr):
        """ Internal function; reads 52 bytes; returns EnergyCenter dict """
        # Called EnergyNano in OpenMIMS
        # Added b field index, energy center enabled, and frequency from MaskNano
        d = {}
        d['detector'], d['start'], d['step size'], d['center'], \
            d['delta'], d['count time'], d['b field index'], \
            d['energy center enabled'], d['frequency'] = \
            unpack(self._bo + '3i 4x 2d i 4x 3i', hdr.read(52))

        d['energy center enabled'] = bool(d['energy center enabled'])
        d['count time'] /= 100  # 10 ms increments to seconds
        if d['detector'] < 0:
            d['detector'] = None
        if d['b field index'] < 0:
            d['b field index'] = None
        return d

    def _e0s_center(self, hdr):
        """ Internal function; reads 40 bytes; returns E0sCenter dict """
        # Called E0SNano in OpenMIMS
        # b field index and e0s center enabled added to sub dict from main nano header
        d = {}
        d['b field index'], d['detector'], d['start'], d['step size'], \
            d['count time'], d['center'], d['80% width'], d['E0S center enabled'] = \
            unpack(self._bo + '5i 2d i', hdr.read(40))

        d['E0S center enabled'] = bool(d['E0S center enabled'])
        d['count time'] /= 100  # 10 ms increments to seconds
        if d['detector'] < 0:
            d['detector'] = None
        if d['b field index'] < 0:
            d['b field index'] = None
        return d

    def _bfield(self, hdr):
        """ Internal function; reads 2840 bytes; returns BField dict """
        # Called TabBFieldNano in OpenMIMS
        d = {}
        d['b field enabled'], d['b field'], d['wait time'], \
            d['time per pixel'], d['time per step'], d['wait time computed'], \
            d['E0W offset'], d['Q'], d['LF4'], d['hex val'], d['frames per bfield'] = \
            unpack(self._bo + '4i d 6i', hdr.read(48))

        d['b field enabled'] = bool(d['b field enabled'])
        d['wait time computed'] = bool(d['wait time computed'])
        d['wait time'] = d['wait time']/1e6
        d['time per pixel'] = d['time per pixel']/1e6

        # 8 bytes unused
        hdr.seek(8, 1)

        # There appear to be 12 trolleys stored.
        # The following labels are true for NS50L and file version 4108.
        # Anywhere else different? What are labels for missing?
        # idx  trolley  idx in header (if all enabled)
        # 0    FCs?     -2
        # 1    T1        0
        # 2    T2        1
        # 3    T3        2
        # 4    T4        3
        # 5    T5        4
        # 6    ?        -3
        # 7    ?        -2
        # 8    SE       -1
        # 9    ?        -3
        # 10   T6        5
        # 11   T7        6
        trolleys = []
        for t in range(12):
            trolleys.append(self._trolley(hdr))

        for t in range(12):
            trolleys[t].update(self._phd(hdr))

        # Add detector index that links trolley to detector and
        # trolley names. Don't know how to do this for EMBig, LD etc.
        for t in range(12):
            if t in (1,2,3,4,5):
                    trolleys[t]['trolley label'] = 'Trolley {}'.format(t)
                    trolleys[t]['detector label'] = 'Detector {}'.format(t)
            elif t in (10,11):
                    trolleys[t]['trolley label'] = 'Trolley {}'.format(t - 4)
                    trolleys[t]['detector label'] = 'Detector {}'.format(t - 4)
            elif t == 8:
                trolleys[t]['trolley label'] = 'SE'
                trolleys[t]['detector label'] = 'SE'
            else:
                trolleys[t]['trolley label'] = 'non-trolley {}'.format(t)
                trolleys[t]['detector label'] = ''

        d['Trolleys'] = trolleys
        return d

    def _trolley(self, hdr):
        """ Internal function; reads 192 or 208 bytes; returns Trolley dict """
        # Called TabTrolleyNano in OpenMIMS
        d = {}
        # exit slit seems to be incorrect
        d['label'], d['mass'], d['radius'], d['deflection plate 1'], \
            d['deflection plate 2'], d['detector'], d['exit slit'], d['real trolley'], \
            d['cameca trolley index'], d['peakcenter index'], d['peakcenter follow'], d['focus'], \
            d['hmr start'], d['start dac plate 1'], d['start dac plate 2'], \
            d['hmr step'], d['hmr points'], d['hmr count time'], \
            d['used for baseline'], d['50% width'], d['peakcenter side'], \
            d['peakcenter count time'], d['used for sib center'], d['unit correction'], \
            d['deflection'], d['used for energy center'], d['used for E0S center'] = \
            unpack(self._bo + '64s 2d 8i 2d 6i d 4i d 2i', hdr.read(192))

        # 16 extra bytes per trolley entry, not in OpenMIMS
        # Only certain versions?
        if self.header['file version'] >= 4108:
            hdr.seek(16, 1)

        # Cleanup
        d['label'] = self._cleanup_string(d['label'])
        d['used for baseline'] = bool(d['used for baseline'])
        d['used for sib center'] = bool(d['used for sib center'])
        d['used for energy center'] = bool(d['used for energy center'])
        d['used for E0S center'] = bool(d['used for E0S center'])
        d['real trolley'] = bool(d['real trolley'])
        d['peakcenter side'] = _peakcenter_sides.get(d['peakcenter side'],
                                                         str(d['peakcenter side']))
        d['detector'] = _detectors.get(d['detector'], str(d['detector']))
        d['hmr count time'] /= 100
        d['peakcenter count time'] /= 100

        # If trolley is real and index is >= 0, then it is enabled. If it is real and
        # index is -1, it is not enabled. If index is < -1, it should also be not real.
        d['trolley enabled'] = bool(d['real trolley'] and d['cameca trolley index'] >= 0)

        return d

    def _phd(self, hdr):
        """ Internal function; reads 24 bytes; returns Phd dict """
        # Called PHDTrolleyNano in OpenMIMS
        d = {}
        d['used for phd scan'], d['phd start'], d['phd step size'], \
            d['phd points'], d['phd count time'], d['phd scan repeat'] = \
            unpack(self._bo + '6i', hdr.read(24))

        d['used for phd scan'] = bool(d['used for phd scan'])
        d['phd count time'] /= 100

        # 24 extra bytes per phd entry, not in OpenMIMS
        # Only certain versions?
        if self.header['file version'] >= 4108:
            hdr.seek(24, 1)

        return d

    def _primary_beam(self, hdr):
        """ Internal function; reads 552 bytes; returns PrimaryBeam dict """
        # Called ApPrimaryNano in OpenMIMS
        d = {}
        start_position = hdr.tell()
        d['source'], d['current start'], d['current end'], d['Lduo'], d['L1'] =\
            unpack(self._bo + '8s 4i', hdr.read(24))

        # Each widths list is 10 ints long
        d['Dduo'] = unpack(self._bo + 'i', hdr.read(4))[0]
        d['Dduo widths'] = tuple(unpack(self._bo + '10i', hdr.read(40)))
        d['D0'] = unpack(self._bo + 'i', hdr.read(4))[0]
        d['D0 widths'] = tuple(unpack(self._bo + '10i', hdr.read(40)))
        d['D1'] = unpack(self._bo + 'i', hdr.read(4))[0]
        d['D1 widths'] = tuple(unpack(self._bo + '10i', hdr.read(40)))

        # 4 bytes unused
        hdr.seek(4, 1)
        d['raster'], d['oct45'], d['oct90'], d['E0P'], d['pressure analysis chamber'] = \
            unpack(self._bo + '4d 32s', hdr.read(64))

        d['source'] = self._cleanup_string(d['source'])
        d['pressure analysis chamber'] = self._cleanup_string(d['pressure analysis chamber'])

        if self.header['analysis version'] >= 3:
            d['L0'] = unpack(self._bo + 'i', hdr.read(4))[0]
        if self.header['analysis version'] >= 4:
            d['hv cesium'], d['hv duo'] = unpack(self._bo + '2i', hdr.read(8))
            # DCs not in OpenMIMS; only in certain release/version?
            d['Dcs'] = unpack(self._bo + 'i', hdr.read(4))[0]
            d['Dcs widths'] = tuple(unpack(self._bo + '10i', hdr.read(40)))

        # skip bytes until total read in this function is 552
        # OpenMIMS: size_Ap_primary_nano = 552
        # Newer versions have rest filled with \xCC continuation bytes, but
        # older versions have null-bytes, but not all bytes are null!!
        # The numbers do not seem to represent anything, though, so can be skipped.
        hdr.seek(start_position + 552)
        return d

    def _secondary_beam(self, hdr):
        """ Internal function; reads 192 bytes; returns SecondaryBeam dict. """
        # Called ApSecondaryNano in OpenMIMS
        d = {}
        tmp = unpack(self._bo + 'd 42i 2d', hdr.read(192))
        d['E0W'], d['ES'] = tmp[:2]
        d['ES widths'] = tmp[2:12]
        d['ES heights'] = tuple(tmp[12:22])
        d['AS'] = tmp[22]
        d['AS widths'] = tuple(tmp[23:33])
        d['AS heights'] = tuple(tmp[33:43])
        d['EnS'], d['EnS width'] = tmp[43:]
        return d

    def _detectors1(self, hdr):
        """ Internal function; reads 808 bytes, returns Detectors dict, part 1. """
        d = {}
        d['FCs'] = self._exit_slits(hdr)

        for n in range(1, 6):
            det = 'Detector {}'.format(n)
            d[det] = self._exit_slits(hdr)

        d['LD'] = {}
        d['LD']['exit slit width'], d['LD']['exit slit coeff a'], \
            d['LD']['exit slit coeff b'], d['E0S'], \
            d['pressure multicollection chamber'], \
            d['FCs']['fc background setup positive'], \
            d['FCs']['fc background setup negative'] = \
            unpack(self._bo + '4d 32s 2i', hdr.read(72))

        d['pressure multicollection chamber'] = \
            self._cleanup_string(d['pressure multicollection chamber'])

        for n in range(1, 6):
            det = 'Detector {}'.format(n)
            d[det].update(self._electron_multiplier(hdr))

        d['LD'].update(self._electron_multiplier(hdr))

        d['EMBig'] = self._exit_slits(hdr)
        d['EMBig'].update(self._electron_multiplier(hdr))

        # 8 bytes unused
        hdr.seek(8, 1)
        return d

    def _detectors2(self, hdr):
        """ Internal function; reads 488 bytes, returns Detectors dict, part 2. """
        # Called AnalysisParam in OpenMIMS, first part only
        # presets separate, last part in _detectors3
        d = {}
        d['Detector 6'] = self._exit_slits(hdr)
        d['Detector 6'].update(self._electron_multiplier(hdr))
        d['Detector 7'] = self._exit_slits(hdr)
        d['Detector 7'].update(self._electron_multiplier(hdr))
        d['exit slit xl'] = unpack(self._bo + '70i', hdr.read(280))
        return d

    def _detectors3(self, hdr):
        """ Internal function; reads 100 bytes, returns Detectors dict, part 3. """
        # Called AnalysisParam in OpenMIMS, only last part
        d = {}
        d['TIC'] = self._electron_multiplier(hdr)

        for n in range(1, 8):
            det = 'Detector {}'.format(n)
            d[det] = {}
            d[det]['fc background setup positive'], d[det]['fc background setup negative'] = \
                unpack(self._bo + '2i', hdr.read(8))

        for n in range(1, 8):
            det = 'Detector {}'.format(n)
            det_type = unpack(self._bo + 'i', hdr.read(4))[0]
            d[det]['detector'] = _detectors.get(det_type, str(det_type))
        return d

    def _exit_slits(self, hdr):
        """ Internal function; reads 88 bytes, returns exit slit dict. """
        # Does not exist separately in OpenMIMS, part of ApSecondaryNano and AnalysisParam
        d = {}
        # Each detector exit slit has:
        # - a position (0, 1, 2)
        # - a size (normal, large, xl)
        # The exit slits widths (and heights) are a 3x5 matrix where coordinate (size, pos)
        #  returns actual width (height). positions 4 and 5 are 0 (for future expansion?)
        # Size XL not stored in same part of header, and only in analysis version >= 5, so we
        # return a list of length 5 with 0s here. Slits 0, 1, 2 are called slit 1, slit 2,
        # slit 3, so add labels to avoid confusion.

        d['exit slit'], d['exit slit size'] = \
            unpack(self._bo + '2i', hdr.read(8))
        d['exit slit label'] = _exit_slit_labels.get(d['exit slit'], str(d['exit slit']))
        d['exit slit size label'] = \
            _exit_slit_size_labels.get(d['exit slit size'], str(d['exit slit size']))

        w0 = tuple(unpack(self._bo + '5i', hdr.read(20)))
        w1 = tuple(unpack(self._bo + '5i', hdr.read(20)))
        w2 = (0,0,0,0,0)
        d['exit slit widths'] = (w0, w1, w2)
        h0 = tuple(unpack(self._bo + '5i', hdr.read(20)))
        h1 = tuple(unpack(self._bo + '5i', hdr.read(20)))
        h2 = (0,0,0,0,0)
        d['exit slit heights'] = (h0, h1, h2)
        return d

    def _electron_multiplier(self, hdr):
        """ Internal function; reads 16 bytes, returns EM dict. """
        d = {}
        d['em yield'], d['em background'], d['em deadtime'] = \
            unpack(self._bo + 'd 2i', hdr.read(16))
        return d

    def _sims_header(self, hdr):
        """Internal function, reads 1240 bytes, returns SIMSHeader dict. """
        # Called DefAnalysisBis and DefEps in OpenMIMS
        d = {}
        d['simsheader version'], d['original filename'], d['matrix'], \
            d['sigref auto'], d['sigref points'], d['sigref delta'], \
            d['sigref scan time'], d['sigref measure time'], \
            d['sigref beam on time'], d['eps centering enabled'], d['eps enabled'], \
            d['eps central energy'], d['eps b field'] = \
            unpack(self._bo + 'i 256s 256s 10i', hdr.read(556))

        d['EPSCentralSpecies'] = self._species(hdr)
        d['EPSReferenceSpecies'] = self._species(hdr)

        # Don't know how long method name is, runs into null-padded zone.
        d['eps ref mass tube hv'], d['eps ref mass tube hv max var'], \
            d['sample rotation'], d['sample rotation speed'], \
            d['sample rotation synced'], d['sample name'], \
            d['user name'], d['method name'] = \
            unpack(self._bo + '2d 3i 80s 32s 256s', hdr.read(396))

        d['original filename'] = self._cleanup_string(d['original filename'])
        d['matrix'] = self._cleanup_string(d['matrix'])
        d['sample name'] = self._cleanup_string(d['sample name'])
        d['user name'] = self._cleanup_string(d['user name'])
        d['method name'] = self._cleanup_string(d['method name'])

        d['sigref auto'] = bool(d['sigref auto'])
        d['eps centering enabled'] = bool(d['eps centering enabled'])
        d['eps enabled'] = bool(d['eps enabled'])
        d['sample rotation'] = bool(d['sample rotation'])
        d['sample rotation synced'] = bool(d['sample rotation synced'])
        d['sigref scan time'] /= 10  # 0.1 sec increments
        return d

    def _preset_start(self, hdr):
        """ Internal function; read 0 bytes (8 and back up),
            returns True or False detecting start of preset.
        """
        test = hdr.read(8)
        hdr.seek(-8, 1)

        try:
            test = self._cleanup_string(test)
        except UnicodeDecodeError:
            # Some non-null filler bytes
            return False

        # First entry in preset is .isf filename with full path
        # If preset is used at all, first entry is non-null.
        # Paths start with / (older Sun systems), or drive letter (D: newer Windows systems)
        if re.match('[A-Z]:', test) or re.match('/.', test):
            return True
        else:
            return False

    def _preset(self, hdr, group=None):
        """ Internal function; reads 1080 (slits) or 3640 (lenses) bytes,
            returns a Preset dict.
        """
        # Called ApPresetSlit/ApPresetLens in OpenMIMS
        # ApPresetDef and ApParamPreset combined here.
        if not group:
            raise ValueError("Group not set, select either 'slit' or 'lens'.")

        d = {}
        start_position = hdr.tell()
        d['isf filename'], d['preset name'], d['calibration date'], \
            d['enabled'], d['parameters'] = \
            unpack(self._bo + '256s 224s 32s 2i', hdr.read(520))

        d['enabled'] = bool(d['enabled'])
        d['isf filename'] = self._cleanup_string(d['isf filename'])
        d['preset name'] = self._cleanup_string(d['preset name'])
        d['calibration date'] = self._cleanup_string(d['calibration date'])
        d['calibration date'] = self._cleanup_date(d['calibration date'])

        # Presets have a fixed length: 1080 for slits, 3640 for lenses.
        # Padded with null- or CC bytes (but there may be other stuff in there).
        # There are more than d['parameters'] parameters in here, but
        # they seem to be "left-overs" from previous presets. Much the
        # same as strings which have more text after the terminating null-byte.
        # Only read first d['parameters'] parameters.
        for p in range(d['parameters']):
            param_id, value, param_name = \
                unpack(self._bo + '2i 20s', hdr.read(28))

            param_name = self._cleanup_string(param_name)
            if not param_name:
                param_name = str(param_id)
            d[param_name] = value

        current_position = hdr.tell()
        if group == 'slit':
            skip = 1080 - (current_position - start_position)
        else:
            skip = 3640 - (current_position - start_position)
        hdr.seek(skip, 1)
        return d

    def _presets(self, hdr):
        """ Internal function; reads 11400 bytes, returns Presets dict. """
        # presput/slit = 1080
        # presput/lens = 3640
        # measure/slit = 1080
        # measure/lens = 3640
        # 2 x 1080 padding
        # padding can be before presput, inbetween presput and measure, and after measure.

        d = {}
        d['Presputter'] = {}
        padding = 0
        if not self._preset_start(hdr):
            hdr.seek(1080, 1)
            padding += 1
        d['Presputter']['Slits'] = self._preset(hdr, group='slit')
        d['Presputter']['Lenses'] = self._preset(hdr, group='lens')
        d['Measure'] = {}
        if not self._preset_start(hdr):
            hdr.seek(1080, 1)
            padding += 1
        d['Measure']['Slits'] = self._preset(hdr, group='slit')
        d['Measure']['Lenses'] = self._preset(hdr, group='lens')
        hdr.seek(1080 * (2 - padding), 1)
        return d

    def _image_hdr(self, hdr):
        """ Internal function; reads 84 bytes, returns Image dict. """
        # Called ... in OpenMIMS
        d = {}
        d['header size'], d['type'], d['width'], d['height'], d['bytes per pixel'], \
            d['masses'], d['planes'], d['raster'], d['original filename'] = \
            unpack(self._bo + 'i 6h i 64s', hdr.read(84))

        # Called nickname in OpenMIMS
        d['original filename'] = self._cleanup_string(d['original filename'])
        if d['header size'] != 84:
            raise ValueError("Image header size is {}, not 84.".format(d['header size']))
        return d

    def _isotopes_hdr(self, hdr):
        """ Internal function; reads 176 bytes, returns Isotopes dict. """
        # Not in OpenMIMS
        d = {}
        d['blocks'], d['frames per block'], d['rejection sigma'], ratios = \
            unpack(self._bo + '4i', hdr.read(16))
        # ratios is the number of ratios to follow. Each ratio is a set of two ints.
        # Each int is the index (0-index) of the species in the mass list. First int
        # is numerator, second is denomenator of ratio.
        r = unpack(self._bo + '{}i'.format(2*ratios), hdr.read(2*4*ratios))
        rtxt = tuple(self.header['label list'][n] for n in r)
        rfmt = tuple(self.header['label list fmt'][n] for n in r)

        d['ratios index'] = tuple((r[n], r[n+1]) for n in range(0, 2*ratios, 2))
        d['ratios'] = tuple((rtxt[n], rtxt[n+1]) for n in range(0, 2*ratios, 2))
        d['ratios fmt'] = tuple('{}\slash {}'.format(rfmt[n], rfmt[n+1]) for n in range(0, 2*ratios, 2))
        # rest is filler with \xFF
        hdr.seek(176 - 16 - 2*4*ratios, 1)
        return d

    def _image_data(self):
        """ internal function; read data for image type. """
        if self.header['Image']['bytes per pixel'] == 2:
            # 16-bit unsigned integers, short
            dt = np.dtype(self._bo + 'H')
        elif self.header['Image']['bytes per pixel'] == 4:
            # 32-bit unsigned integers, int
            dt = np.dtype(self._bo + 'I')

        shape = [self.header['Image']['planes'],
                 self.header['Image']['masses'],
                 self.header['Image']['height'],
                 self.header['Image']['width']]

        self.fh.seek(self.header['header size'])

        compressedfiles = (gzip.GzipFile, bz2.BZ2File, tarfile.ExFileObject, io.BytesIO)
        if lzma:
            compressedfiles += (lzma.LZMAFile,)

        if isinstance(self.fh, compressedfiles):
            self.data = np.fromstring(self.fh.read(), dtype=dt).reshape(shape)
        else:
            self.data = np.fromfile(self.fh, dtype=dt).reshape(shape)

        # We want to have a cube of contiguous data (stacked images) for each
        # mass. Swap axes 0 and 1. Returns a view, so make full copy to make
        # data access faster.
        self.data = self.data.swapaxes(0, 1).copy()

        if pd:
            self.data = pd.Panel4D(self.data, labels=self.header['label list'])
            self.data.labels.name = 'species'
            self.data.items.name = 'frame'
            self.data.major_axis.name = 'y'
            self.data.minor_axis.name = 'x'

    def _isotope_data(self):
        """ Internal function; read data from .is or .dp file. """
        # Data structure:
        #   header
        #   1 int (M): no. blocks (i.e. masses)
        #   M blocks:
        #     (each block)
        #     1 int (N): no. points (i.e. frames)
        #     N doubles: cumulative count time in s
        #     N doubles: data
        self.fh.seek(self.header['header size'])
        blocks = unpack(self._bo + 'i', self.fh.read(4))[0]

        data = []
        for block in range(blocks):
            points = unpack(self._bo + 'i', self.fh.read(4))[0]
            d = np.fromfile(self.fh, dtype=self._bo+'f8', count=2*points).reshape(2, points)
            data.append(d[1])

        if pd:
            self._data_corr = pd.DataFrame(data, index=self.header['label list']).T
            self._data_corr.index.name = 'frame'
            self._data_corr.columns.name = 'counts/s'
        else:
            self._data_corr = np.vstack(data)

    def _isotope_txt_data(self):
        """ Internal function; read data from .is_txt or .dp_txt file. """
        txt = None
        with open(self.filename + '_txt', mode='rt') as fh:
            txt = fh.readlines()

        data = []
        frames = self.header['frames']
        l = 0
        while l < len(txt):
            if txt[l].startswith('B ='):
                Tc = txt[l].split('=')[-1].strip().strip(' ms')
                Tc = float(Tc)/1000
                d = np.loadtxt(txt[l+2 : l+2+frames])
                data.append(d[:,1]/Tc)
                l += 2 + frames
            l += 1

        if pd:
            self.data = pd.DataFrame(data, index=self.header['label list']).T
            self.data.index.name = 'frame'
            self.data.columns.name = 'counts/s'
        else:
            self.data = np.vstack(data)

    def _read_chk_is(self):
        """ Internal function, reads .chk_is file, extracts background calibration. """
        fname = os.path.splitext(self.filename)[0] + '.chk_is'
        table = []
        bg_before = []
        bg_after = []
        with open(fname, mode='rt') as fh:
            for line in fh:
                if line.startswith('FC Background before acq'):
                    bg_before = line.split(':')[1].strip().split('=')
                elif line.startswith('FC Background after acq'):
                    bg_after = line.split(':')[1].strip().split('=')
                elif line.startswith('|'):
                    table.append(line)

        # Parse analysis background
        for n in range(0, len(bg_before), 2):
            det = bg_before[n].replace('Det', 'Detector ').strip()
            try:
                bg = float(bg_before[n+1].strip())
            except ValueError:
                bg = 0
            self.header['Detectors'][det]['fc background before analysis'] = bg

        for n in range(0, len(bg_after), 2):
            det = bg_after[n].replace('Det', 'Detector ').strip()
            try:
                bg = float(bg_after[n+1].strip())
            except ValueError:
                bg = 0
            self.header['Detectors'][det]['fc background after analysis'] = bg

        # Parse baseline background if found
        if table:
            background = table[0].strip().strip('|').split('|')
            background = [float(b.strip()) for b in background[1:]]
            idx = table[2].strip().strip('|').split('|')
            idx = [int(i.strip('Mas# ')) for i in idx[2:]]
            species = [self.header['label list'][i-1] for i in idx]
            baseline = dict(zip(species, background))

            for spc, mt in self.header['MassTable'].items():
                if spc in baseline.keys():
                    bfi = mt['b field index']
                    tri = mt['trolley index']
                    tr = self.header['BFields'][bfi]['Trolleys'][tri]
                    if tr['detector'] == 'EM':
                        tr['em background baseline'] = baseline[spc]
                    elif tr['detector'] == 'FC':
                        tr['fc background baseline'] = baseline[spc]

    def _beamstability_data(self):
        """ Internal function; read data in .bs beam stability file."""
        traces = unpack(self._bo + 'i', self.fh.read(4))[0]
        time = []
        data = []
        for t in range(traces):
            points = unpack(self._bo + 'i', self.fh.read(4))[0]
            d = np.fromfile(self.fh, dtype=self._bo+'f8', count=2*points).reshape(2, points)
            time = d[0]
            data.append(d[1])

        if pd:
            self.data = pd.DataFrame(data).T
            self.data.index = time
            self.data.index.name = 'seconds'
            self.data.columns = self.header['label list']
            self.data.columns.name = 'counts/s'
        else:
            self._data_corr = np.vstack(time, data)


    def _cleanup_string(self, bytes):
        """ Internal function; cuts off bytes at first null-byte,
            decodes bytes as latin-1 string, returns string
        """
        try:
            b = bytes.index(b'\x00')
        except ValueError:
            return bytes.decode('latin-1').strip()
        else:
            return bytes[:b].decode('latin-1').strip()

    def _cleanup_date(self, date):
        """ Internal function; reads date-string, returns Python datetime object.
            Assumes date-part and time-part are space separated, date is dot-separated,
            and time is colon-separated. Returns None if date is empty, contains 'N/A',
            or is not a string.
        """
        if (not date
            or not isinstance(date, str)
            or 'N/A' in date):
            return None

        date, time = date.split()
        day, month, year = date.split('.')
        hour, minute = time.split(':')
        year, month, day, hour, minute = [int(x) for x in (year, month, day, hour, minute)]

        # For 2-digit years, 1969/2068 is the wrap-around (POSIX standard)
        if (69 <= year < 100):
            year += 1900
        elif (0 <= year < 69):
            year += 2000

        return datetime.datetime(year, month, day, hour, minute)

    def _chomp(self, hdr, filler=(b'\x00\x00\x00\x00', b'\xCC\xCC\xCC\xCC'), chunk=4):
        """ Internal function.

            Reads and discards filler bytes, by default null (\\x00) and continue (\\xCC) bytes.
            Stops at start of first non-filler byte. Reads chunk bytes at the time.
        """
        filler = tuple(filler)
        bytes = hdr.read(chunk)
        while(bytes and bytes in filler):
            bytes = hdr.read(chunk)
        hdr.seek(-1 * chunk, 1)


class SIMS(SIMSReader, TransparentOpen):
    """ Read a (nano)SIMS file and load the full header and image data. """
    def __init__(self, filename, file_in_archive=0, password=None):
        """ Create a SIMS object that will hold all the header information and image data.

            Usage: s = sims.SIMS('filename.im' | 'filename.im.bz2' | fileobject)

            Header information is stored as a nested Python dict in SIMS.header,
            while data is stored in SIMS.data as a pandas.Panel4D object if
            pandas is installed or a numpy ndarray otherwise.

            This class can open Cameca (nano)SIMS files and transparently supports
            compressed files (gzip, bzip2, xz, lzma, zip, 7zip) and opening from
            multifile archives (tar, compressed tar, zip, 7zip). Set file_in_archive
            to the filename to extract, or the sequence number of the file in the
            archive (0 is the first file). For encrypted archives (zip, 7z) set
            password to access the data. For zip format, password must be a
            byte-string.

            It's also possible to supply a file object to an already opened file.
            In fact, SIMS can read from anything that provides a read() function,
            although reading from a buffered object (with seek() and tell()
            support) is much more efficient.

            SIMS supports the 'with' statement.
        """
        if not filename:
            return

        TransparentOpen.__init__(self, filename, file_in_archive=file_in_archive,
                                 password=password)
        self.fh.seek(0)
        SIMSReader.__init__(self, self.fh, filename=filename)

        self.peek()
        self.read_header()
        self.read_data()
        self.close()


#####################################################################################
# Not implemented yet in read_header
#
#     class CalCond
#         int n_delta;
#         int np_delta;
#         int tps_comptage;
#         int nb_cycles;
#         double no_used2;
#         double cal_ref_mass;
#         String symbol;



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Convert and load Cameca and L'image look-up tables.

    Module to load Look-Up Tables (LUTs) from the Cameca-supplied lut
    files and the L'image colour table file. Loaded LUTs are converted to
    Matplotlib Colormaps, and registered with Matplotlib. Use mpl.colormaps()
    to see all registered Colormaps and mpl.set_cmap() to set the Colormap
    to use.

    The Cameca and L'image LUTs are distributed with the sims module. Cameca .lut
    files are located in the LUT directory which is stored as sims.lut.lut_dir.
    The L'image Colour Table file is located in the same directory and is stored
    as sims.lut.limage_file. Both can be overriden to load from a different
    directory.
"""
import matplotlib.pyplot as mpl
import matplotlib.colors as mpc
import os, pkg_resources

__all__ = ['load_cameca_lut', 'load_limage_lut']

lut_dir = pkg_resources.resource_filename(__name__, 'lut')
limage_file = os.path.join(lut_dir, 'limagecolors.tbl')

def load_cameca_lut(*names, smooth=True):
    """ Load a Cameca Look-Up Table.

        Usage: load_cameca_lut('cameca bw', 'temp', ..., smooth=True)

        Reads one or more LUTs from the Cameca supplied .lut files, converts it to a
        Matplotlib Colormap and registers it with Matplotlib. The Colormaps will
        be registered with a 'cameca ' prefix to distinguish them from the L'image
        LUTs and Matplotlib's own Colormaps. To (re)load a specific LUT, the
        'cameca ' prefix may be omitted. If no name is given, all LUTs in the lut-
        directory will be loaded.

        By default, a smooth gradient Colormap (LinearSegmentedColormap) is created
        from the data in the LUT. Some LUTs, however, are meant to be used with
        hard edges and strong contrast between adjacent colours. To disable the
        smooth gradient, set smooth=False and a ListedColormap will be created in
        stead.
    """
    if len(names) == 0:
        fnames = os.listdir(lut_dir)
        fnames = [os.path.join(lut_dir, n) for n in fnames if n.endswith('.lut')]
    else:
        names = [n[7:] for n in names if n.startswith('cameca ')]
        fnames = [os.path.join(lut_dir, n, '.lut') for n in names]

    # All luts need to be 0-1 normalized
    norm = mpc.Normalize(vmin=0, vmax=255)

    for fname in fnames:
        with open(fname, mode='rb') as fh:
            # skip header
            fh.seek(76)
            # Data is in single bytes, 3 rows of unknown length (either 128 or 256)
            # Data is also doubled, each row appears twice (from-to?)
            lut_data = np.fromfile(fh, dtype='B').reshape(3, -1).T[::2]

        name = os.path.basename(fname)
        name = 'cameca ' + os.path.splitext(name)[0]

        lut_data = norm(lut_data)

        if smooth:
            lut = mpc.LinearSegmentedColormap.from_list(name, lut_data)
        else:
            lut = mpc.ListedColormap(lut_data, name=name)
        mpl.register_cmap(cmap=lut)


def load_limage_lut(*names, smooth=True):
    """ Load a L'image Look-Up Table.

        Usage: load_limage_lut('blue/white', 'limage prism', ..., smooth=True)

        Reads one or more LUTs from the L'image colortable file, converts it to a
        Matplotlib Colormap and registers it with Matplotlib. The Colormaps will
        be registered with a 'limage ' prefix to distinguish them from the Cameca-
        supplied LUTs and Matplotlib's own Colormaps. To (re)load a specific LUT,
        the 'limage ' prefix may be omitted. If no name is given, all LUTs in the
        file will be loaded.

        By default, a smooth gradient Colormap (LinearSegmentedColormap) is created
        from the data in the LUT. Some LUTs, however, are meant to be used with
        hard edges and strong contrast between adjacent colours. To disable the
        smooth gradient, set smooth=False and a ListedColormap will be created in
        stead.
    """
    names = [n[7:] for n in names if n.startswith('limage ')]

    fh = open(limage_file, mode='rb')
    hdr = fh.read(32)
    data = fh.read()
    fh.close()

    if 'PV-WAVE CT' not in hdr[:16].decode('utf-8'):
        raise TypeError('File {} is not a PV-Wave ColorTable.'.format(limage_file))

    # All luts need to be 0-1 normalized
    norm = mpc.Normalize(vmin=0, vmax=255)

    # Number of LUTs stored as number in an ascii-encoded, null-padded string (jeez)
    n_luts = hdr[16:].decode('utf-8').strip(' \x00')
    n_luts = int(n_luts)
    offset = 32 * n_luts
    for l in range(n_luts):
        name = data[l*32:(l+1)*32].decode('utf-8').strip(' \x00').lower()
        # len(name) == 0 -> load all
        # lut_name in names -> this is (the) one we need
        # not name.startswith('?') -> signifies empty table entry
        if (not name.startswith('?')) and ((name in names) or (len(names) == 0)):
            name = 'limage ' + name

            lut_data = np.fromstring(data[offset + l*768:offset + (l+1)*768],
                                     dtype='B').reshape(3, 256).T
            lut_data = norm(lut_data)

            if smooth:
                lut = mpc.LinearSegmentedColormap.from_list(name, lut_data)
            else:
                lut = mpc.ListedColormap(lut_data, name=name)

            mpl.register_cmap(cmap=lut)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Tools and utilities for the sims module. """
from __future__ import print_function, division
import re
import os
import io
import json
import warnings
import datetime
import numpy as np
import pandas as pd
from skimage.feature import register_translation
from scipy.ndimage import shift
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
import matplotlib.pyplot as mpl

__all__ = [
    'coulomb',
    'ions_per_amp',
    'format_species',
    'thumbnails',
    'coordinates',
    'align',
    'em_correct',
    'fc_correct',
    'export_fits',
    'export_header',
    'export_matlab'
]

# Factor for converting A to counts/s on faraday cups.
# Cameca uses 6.24142e18, which is incorrect rounding: 6.241509... e18
# Number used here is NIST 2010CODATA
# http://physics.nist.gov/cgi-bin/cuu/Value?e
coulomb = 1.6021766208e-19  # charge/ion
ions_per_amp = 1/coulomb  # ions/A

def format_species(name, mhchem=False, mathrm=False):
    """ Format the name of a chemical species.

        Usage: formatted_name = format_species(name, mhchem=False)

        Takes a string which represents an atomic or molecular species
        and returns a string with LaTeX-style sub- and superscripts for input
        into Matplotlib. Multiple atoms in the input string are expected to be
        space-separated within a molecule, as in Cameca formatting. Irregularly
        formatted strings are silently skipped.

        If mhchem=True, then the \ce command from the mhchem package is used to
        format species names. This gives much better results, but requires LaTeX
        and mhchem to be installed. For this to typeset properly in matplotlib,
        set 'text.usetex' to True and include '\\usepackage{mhchem}' in
        'text.latex.preamble' (or 'pgf.preamble' when exporting as pgf) in
        rcParams or matplotlibrc.

        If mathrm=True, a \mathrm command is inserted, to typeset upright
        letters. Useful when the full LaTeX engine is used (text.usetex: True in
        matplotlibrc); LaTeX typesets math mode text in italic by default. This
        option is ignored if mhchem=True.

        Example:
        >>> format_species('12C2 2H')
        '${}^{12}C_{2}{}^{2}H$'

        >>> format_species('12C2 2H', mathrm=True)
        '$\mathrm{{}^{12}C_{2}{}^{2}H}$'

        >>> format_species('12C2 2H', mhchem=True)
        '\ce{{}^{12}C_{2}{}^{2}H}
    """
    # {} is format subst. {{}} is literal {} after expansion.
    # {{{}}} is a subst inside a literal {} after expansion.
    # First {{}} (expands to {}) aligns superscript with following
    # character, not previous.
    # The string 1A2 expands to: '{}^{1}A_{2}'
    mass_tmpl = '{{}}^{{{mass}}}'
    elem_tmpl = '{elem}'
    stoich_tmpl = '_{{{stoich}}}'
    charge_tmpl = '^{{{charge}}}'

    if mhchem:
        begin = '\ce{'
        end = '}'
    elif mathrm:
        begin = '$\mathrm{'
        end = '}$'
    else:
        begin = end = '$'

    body = ''
    for atom in name.split():
        if '+' in atom or '-' in atom:
            body += charge_tmpl.format(charge=atom)
            continue

        parts = re.split('([a-zA-Z]+)', atom)
        if not len(parts) == 3:
            continue

        if not parts[1]:
            continue
        else:
            elem = elem_tmpl.format(elem=parts[1])

        if parts[0]:
            mass = mass_tmpl.format(mass=parts[0])
        else:
            mass = ''

        if parts[2]:
            stoich = stoich_tmpl.format(stoich=parts[2])
        else:
            stoich = ''

        body += mass + elem + stoich

    return begin + body + end


def thumbnails(data, cycle=0, mass=None, labels=None):
    """ Generate a thumbnail sheet. """
    if not mass: mass = range(data.shape[0])
    mass = list(mass)

    row = int(np.floor(np.sqrt(len(mass))))
    diff = np.sqrt(len(mass)) - row
    if diff == 0:
        # square, MxM
        col = row
    elif diff < 0.5:
        # not square, MxN
        col = row + 1
    else:
        # next square, NxN
        row += 1
        col = row

    fig = figure(figsize=(14.2222, 14.2222), dpi=72, facecolor='white')
    for m, n in zip(mass, range(len(mass))):
            ax = mpl.subplot(row, col, n+1)
            ax.axis('off')
            ax.imshow(data[m, cycle])
            if labels: axim.title(labels[n])
            # mpl.colorbar()

    mpl.show()

def coordinates(filelist, **kwargs):
    """ Find all coordinates in a list of image files.

        Usage: fig = sims.coordinates([a.im, b.im], labels=['A', 'B'])

        For each image in the list, the stage coordinates and raster size are extracted.
        A box scaled to the rastersize is plotted for each image on a (X,Y) grid.
        A label for each file can be given. If it's omitted, the filename will be printed
        over the box, but no attempt is made to make the filename fit in the box.

        Returns a matplotlib figure instance.
    """
    labels = kwargs.pop('labels', filelist)

    patches = []
    x_min = None
    fig = mpl.figure()
    ax = fig.gca()

    for fn, lb in zip(filelist, labels):
        s = sims.SIMSOpener(fn)
        s.peek()
        s.read_header()
        s.close()

        # Remember!! In Cameca-land, X is up-down, Y is left-right. Fix here.
        x = s.header['sample y']
        y = s.header['sample x']

        if (x == 0 and y == 0):
            warnings.warn('Coordinates for file {} are 0,0.'.format(fn))

        if (('PrimaryBeam' in s.header.keys()) and ('raster' in s.header['PrimaryBeam'].keys())):
            raster = s.header['PrimaryBeam']['raster']
        elif (('Image' in s.header.keys()) and ('raster' in s.header['Image'].keys())):
            raster = s.header['Image']['raster']/1000
        else:
            warnings.warn('No raster size in header of file {}.'.format(fn))
            raster = 0

        # Text uses center of image coordinate, box uses lower left.
        ax.text(x, y, lb, ha='center', va='center', fontsize=8)

        x -= raster/2
        y -= raster/2

        # Update data limits, relim() used by autoview does not work with collections (yet)
        if not x_min:
            # first pass
            x_min = x
            x_max = x + raster
            y_min = y
            y_max = y + raster
        else:
            if x < x_min: x_min = x
            if x + raster > x_max: x_max = x + raster
            if y < y_min: y_min = y
            if y + raster > y_max: y_max = y + raster

        rect = Rectangle((x,y), raster, raster, ec='black', fc='white', fill=False)
        patches.append(rect)

    collection = PatchCollection(patches, match_original=True)
    ax.add_collection(collection)

    # Set view limits
    x_span = x_max - x_min
    y_span = y_max - y_min
    # Remember!! In Cameca-land, X-axis (up-down) is positive downwards,
    # Y-axis (left-right) is positive going right.
    ax.set_xlim(x_min - 0.1*x_span, x_max + 0.1*x_span)
    ax.set_ylim(y_max + 0.1*y_span, y_min - 0.1*y_span)

    # Keep it square
    ax.set_aspect('equal', adjustable='datalim')
    ax.set_xlabel('Stage Y (μm)')
    ax.set_ylabel('Stage X (μm)')
    return fig


def export_fits(data, filename, extend=False, **kwargs):
    """ Export data to a FITS file.

        Data can be pandas data structure or numpy ndarray.
        Filename is the filename the FITS will will be saved to.

        By default, the data structure will be saved as is, i.e.
        a 3D data cube will be saved as such. Set extend=True to
        unroll the outer dimension and save the remaining data
        structures as a list of HDUs (Header Data Units, see
        FITS documentation); a 3D data cube will be saved as a
        list of 2D images.

        Additional arguments are sent to fits.writeto(), see
        astropy.io.fits for more info.
    """
    fits = None
    try:
        from astropy.io import fits
    except ImportError:
        pass

    try:
        import pyfits as fits
    except ImportError:
        pass

    if not fits:
        msg = 'You need to install either pyfits or astropy to be able to export FITS files.'
        raise ImportError(msg)

    if hasattr(data, 'values'):
        # any of the pandas data structures; extract ndarray
        data = data.values

    # FITS supports the following data formats, given as BITPIX in header
    # BITPIX    format
    # 8         unsigned 8-bit integer
    # 16        signed 16-bit integer
    # 32        two's complement 32-bit integer
    # -32       IEEE 32-bit float
    # -64       IEEE 64-bit float

    # PyFITS handles uint16 and uint32 by storing as int16 and int32
    # and setting BZERO offset. int8, float16, and float128 give KeyError.

    # PyFITS will save (u)int64, but is non standard and at least Source Extractor
    # does not recognise it. Downcast if possible, error otherwise. Everything
    # else is handled by PyFITS.

    if data.dtype in (np.uint64, np.int64):
        if (data.min() >= np.iinfo('i4').min and
            data.max() <= np.iinfo('i4').max):
                data = data.astype('i4', copy=False)
        elif (data.min() >= np.iinfo('u4').min and
            data.max() <= np.iinfo('u4').max):
                data = data.astype('u4', copy=False)
        else:
            msg = 'Data is (u)int64 and cannot safely be downcast to (u)int32.'
            raise ValueError(msg)
    elif data.dtype == np.int8:
        if data.min() >= 0:
            data = data.astype('u1', copy=False)
        else:
            data = data.astype('i2', copy=False)
    elif data.dtype == np.float16:
        data = data.astype('f4', copy=False)
    elif data.dtype == np.float128:
        if (data.min() >= np.finfo('f8').min and
            data.max() <= np.finfo('f8').max) :
                data = data.astype('f8', copy=False)
        else:
            msg = 'Data is float128 and cannot safely be downcast to float64.'
            raise ValueError(msg)

    if extend:
        hl = [fits.PrimaryHDU()] + [fits.ImageHDU(n) for n in data]
        hdulist = fits.HDUList(hl)
        hdulist.writeto(filename, **kwargs)
    else:
        hdu = fits.PrimaryHDU(data)
        hdu.writeto(filename, **kwargs)


def align(data, reference_frame=0, upsample_factor=10, center=True, **kwargs):
    """ Perform image alignment on data.

        usage: aligned_data, shifts = align(data)

        Performs sub-pixel image translation registration (image alignment)
        on the data. Each frame of each element is registered seperately.
        skimage.features.register_translation() is used for obtaining the
        shifts and scipy.ndimage.shift() is used to apply the shifts to the
        data. See the documentation of those functions for more information;
        keyword arguments are passed through. By default, the upsample_factor
        is set to 10.

        The data can be given as a pandas Panel4D (sims default) or Panel
        (a single element), or as a numpy array.

        The reference frame is the frame to which all other images in the
        stack are adjusted. By default this is 0, the first frame. A list
        may also be given, in which case a different reference frame for
        each element can be set.

        If center=True (the default), the shifts are centered to the median
        of the shifts to minimize blank edges.

        Returns the aligned data and the shifts.
    """
    was_panel = False
    was_numpy = False
    if isinstance(data, pd.Panel):
        data = pd.Panel4D(data=data.values, labels=['x'], items=data.items,
                          major_axis=data.major_axis, minor_axis=data.minor_axis)
        was_panel = True
    elif isinstance(data, np.ndarray):
        if data.ndim == 3:
            data = pd.Panel4D(data=[data])
        elif data.ndim == 4:
            data = pd.Panel4D(data=data)
        else:
            msg = 'data must be 3 dimensional (one stack of 2D images), or '
            msg += '4 dimentional (multiple stacks of 2D images).'
            raise TypeError(msg)
        was_numpy = True

    if isinstance(reference_frame, int):
        reference_frame = [reference_frame] * data.shape[0]

    # Calculate shifts
    shifts = []
    for lbl, ref in zip(data.labels, reference_frame):
        shifts_per_lbl = []
        for lyr in data.loc[lbl]:
            sh = register_translation(data.loc[lbl, ref],
                                      data.loc[lbl, lyr],
                                      upsample_factor=upsample_factor)[0]
            shifts_per_lbl.append(sh)
        shifts.append(np.vstack(shifts_per_lbl))
    shifts = pd.Panel(shifts, items=data.labels, minor_axis=['y', 'x'])

    # Center shifts
    if center:
        xmedian = shifts.loc[:,:,'x'].median()
        ymedian = shifts.loc[:,:,'y'].median()
        shifts.loc[:,:,'x'] += xmedian
        shifts.loc[:,:,'y'] += ymedian

    # Apply shifts to data
    shifted = pd.Panel4D(np.zeros(data.shape), labels=data.labels)
    for lbl in data.labels:
        for frm in data.items:
            shifted.loc[lbl, frm] = shift(data.loc[lbl, frm],
                                          shifts.loc[lbl, frm])
    if was_panel:
        return (shifted['x'], shifts['x'])
    elif was_numpy:
        return (shifted.values.squeeze(), shifts.values.squeeze())
    else:
        return (shifted, shifts)

def em_correct(simsobj, species=None, deadtime=None, emyield=None, background=None):
    """ Correct data recorded with EMs for several factors.

        Usage: sims.utils.em_correct(s)

        Applies background, yield, and deadtime correction to all species
        in the data that were recorded with electron multipliers (EMs). By
        default, all correction factors are read from the header. To supply
        other values, give the deadtime (in seconds), emyield (in fraction,
        not %), or background (in same units as data, counts or counts/s)
        as keyword arguments.

        For background it is possible to set up a "baseline" measurement,
        although this is more commonly done for FC detectors. These values
        are read from the .chk_is file. Set the background keyword to
        "baseline" to use these values.

        Correction factors can be given either a single value to be used
        for all species in the data, or as a dict with label:value pairs
        to specify different values per species. This dictionary will
        override the default values for that species, the other species in
        the data will still use the values from the header.

        By default, all EM-recorded species in the data will be corrected.
        It is possible to limit that list by specifying the species keyword
        with a list of species labels. The data of any species not in the list
        will not be altered.

        The EM deadtime correction with deadtime t assumes a non-paralyzable
        detector (see Williamson et al, Anal. Chem, 1988, 60, 2198-2203) and
        the correction applied is:

        Ireal = Icounted/(1 - Icounted * t)

        This adjusts the data in the simsobj. It will also set keywords in the
        header to show that corrections have been applied. It is not possible
        to apply the same corrections more than once.
    """
    EMs = {}
    for spc, mt in simsobj.header['MassTable'].items():
        bfi = mt['b field index']
        tri = mt['trolley index']
        tr = simsobj.header['BFields'][bfi]['Trolleys'][tri]
        if tr['trolley enabled'] and tr['detector'] == 'EM':
            det = simsobj.header['Detectors'][tr['detector label']]
            EMs[spc] = {'deadtime': det['em deadtime']/1e9,  # in ns
                        'yield': det['em yield']/100,  # in %
                        'background': det['em background']}
            if background == 'baseline':
                if tr['used for baseline'] and 'em background baseline' in tr.keys():
                    EM[spc]['background'] = tr['em background baseline']

    # Override species
    if species:
        for spc in species:
            if spc not in EMs.keys():
                msg = '{} not in data or not using an EM.'.format(spc)
                raise KeyError(msg)
        for spc in EMs.keys():
            if spc not in species.keys():
                EMs.pop(spc)

    # Override deadtime values.
    if deadtime is None:
        pass
    elif isinstance(deadtime, (int, float)):
        for spc in EMs.keys():
            EMs[spc]['deadtime'] = deadtime
    elif isinstance(deadtime, dict):
        for spc, dt in deadtime.items():
            # raise KeyError if label does not exist in data.
            EMs[spc]
            EMs[spc]['deadtime'] = dt
    else:
        msg = 'deadtime value not understood. Give either a single value, or a '
        msg += 'dictionary {species: value}, or None to use the deadtime value '
        msg += 'from the header (the default).'
        raise TypeError(msg)

    # Override yield values.
    if emyield is None:
        pass
    elif isinstance(emyield, (int, float)):
        for spc in EMs.keys():
            EMs[spc]['yield'] = emyield
    elif isinstance(emyield, dict):
        for spc, yld in emyield.items():
            # raise KeyError if label does not exist in data.
            EMs[spc]
            EMs[spc]['yield'] = yld
    else:
        msg = 'yield value not understood. Give either a single value, or a '
        msg += 'dictionary {species: value}, or None to use the yield value '
        msg += 'from the header (the default).'
        raise TypeError(msg)

    # Override background values.
    if background is None:
        pass
    elif isinstance(background, (int, float)):
        for spc in EMs.keys():
            EMs[spc]['background'] = background
    elif isinstance(background, dict):
        for spc, bg in background.items():
            # raise KeyError if label does not exist in data.
            EMs[spc]
            EMs[spc]['background'] = bg
    else:
        if isinstance(background, str) and background == 'baseline':
            pass
        else:
            msg = 'background value not understood. Give either a single value, or a '
            msg += 'dictionary {species: value}, or the string "baseline" to use the '
            msg += 'values from the .chk_is file from the baseline measurement, or None '
            msg += 'to use the background value from the header (the default).'
            raise TypeError(msg)

    for spc, dct in EMs.items():
        if not simsobj.header['MassTable'][spc]['background corrected']:
            simsobj.data[spc] -= dct['background']
            simsobj.header['MassTable'][spc]['background corrected'] = True
        if not simsobj.header['MassTable'][spc]['yield corrected']:
            simsobj.data[spc] *= dct['yield']
            simsobj.header['MassTable'][spc]['yield corrected'] = True
        if not simsobj.header['MassTable'][spc]['deadtime corrected']:
            simsobj.data[spc] = simsobj.data[spc].div(1 - dct['deadtime']*simsobj.data[spc])
            simsobj.header['MassTable'][spc]['deadtime corrected'] = True

def fc_correct(simsobj, background='setup', species=None, resistor=None, gain=None):
    """ Correct and convert data recorded with FCs.

        Usage: sims.utils.fc_correct(s, resistor={'16O': 10e9, '18O': 100e9})

        Applies background correction to all species in the data that were
        recorded with Faraday cups (FCs) and converts the data from "cps"
        (units of 0.01 mV/s) to real counts/s.

        By default the background values are read from the header, which is
        referred to as the "setup" background values. Alternatively, the
        background values that were recorded before and after the measurement
        for 1 s intervals can be used. This is referred to as the "analysis"
        background and is only available if the corresponding .chk_is file
        is present. A more elaborate background can be measured by setting
        up a "baseline" measurement. These values are also read from the
        .chk_is file. Finally, custom background values can be supplied
        either as a single value for all FCs, or as a dictionary with a
        value for each species that was recorded with an FC (skip EM
        recorded data). Give the values in "cps".

        Faraday cups work by running the collected charge through a large
        resistor and measuring the voltage across that resistor. In the
        nanoSIMS (what about other instruments?) 2 resistors are present
        for each FC: 10 GOhm and 100 GOhm. They are selected by setting a
        jumper in the Finnigan can and selecting the corresponding value in
        Setup. Which resistor is selected is not recorded in the header. By
        default, a guess will be made by comparing the raw data with the
        Cameca-corrected data. See SIMSReader.read_data() for more info. If
        the guess fails, or is incorrect, resistor values can be supplied
        either as a single value (in Ohm: 10e9 or 100e9) or as a dictionary
        with a value for each species recorded with a FC.

        By default, all FC-recorded species in the data will be corrected.
        It is possible to limit the data correction to only a few species
        by specifying the species keyword with a list of labels The data
        of all species not in the list will not be altered.

        This adjusts the data in the simsobj. It will also set keywords in the
        header to show that background correction has been applied. It is not
        possible to apply the correction more than once.
    """
    # Get all the trolleys that are set to FC, link to species label, and get background.
    FCs = {}
    for spc, mt in simsobj.header['MassTable'].items():
        bfi = mt['b field index']
        tri = mt['trolley index']
        tr = simsobj.header['BFields'][bfi]['Trolleys'][tri]
        if tr['trolley enabled'] and tr['detector'] == 'FC':
            FCs[spc] = {}
            det = simsobj.header['Detectors'][tr['detector label']]
            # Always read setup value, override if other is requested.
            if simsobj.header['polarity'] == '+':
                FCs[spc]['background'] = det['fc background setup positive']
            else:
                FCs[spc]['background'] = det['fc background setup negative']
            if background == 'baseline':
                if tr['used for baseline']:
                    FCs[spc]['background'] = tr['fc background baseline']
                else:
                    msg = 'no baseline measurement found for {}, '.format(spc)
                    msg += '({}), using background value from setup.'.format(simsobj.filename)
                    warnings.warn(msg)
            elif background == 'analysis':
                if 'fc background before analysis' in det.keys():
                    before = det['fc background before analysis']
                    after = det['fc background after analysis']
                    FCs[spc]['background'] = (before + after)/2
                else:
                    msg = 'no analysis background found for {}, '.format(spc)
                    msg += '({}), using background value from setup.'.format(simsobj.filename)
                    warnings.warn(msg)

    # Override species
    if species:
        for spc in species:
            if spc not in FCs.keys():
                msg = '{} not in data or not using a FC.'.format(spc)
                raise KeyError(msg)
        for spc in FCs.keys():
            if spc not in species.keys():
                FCs.pop(spc)

    # Override background values.
    if background is None:
        pass
    elif isinstance(background, (float, int)):
        for spc in FCs.keys():
            FCs[spc]['background'] = background
    elif isinstance(background, dict):
        for spc, bg in background.keys():
            # raise KeyError if label does not exist in data.
            FCs[spc]
            FCs[spc]['background'] = bg
    else:
        if isinstance(background, str) and background in ('setup', 'analysis', 'baseline'):
            pass
        else:
            msg = 'background value not understood. Give either a string, one of '
            msg += '"setup", "analysis", or "baseline", or a single value, or a '
            msg += 'dictionary {species: value}.'
            raise TypeError(msg)

    # Resistor values
    if not resistor:
        resistor = _guess_fc_resistors(simsobj)
        for spc in resistor.index:
            FCs[spc]['resistor'] = resistor[spc]
    elif isinstance(resistor, (float, int)):
        for spc in FCs.keys():
            FCs[spc]['resistor'] = resistor
    elif isinstance(resistor, dict):
        for spc, r in resistor.items():
            # raise KeyError if label does not exist in data.
            FCs[spc]
            FCs[spc]['resistor'] = r
    else:
        msg = 'resistor value not understood. Give either a single value, or a '
        msg += 'dictionary {species: value}, or None to have the resistor value '
        msg += 'guessed automatically.'
        raise TypeError(msg)

    for spc, dct in FCs.items():
        if not simsobj.header['MassTable'][spc]['background corrected']:
            simsobj.data[spc] -= dct['background']
            simsobj.header['MassTable'][spc]['background corrected'] = True
            # 1e5 is the magic factor to go from "cps",
            # which is aparently 0.01 mV/s, to V/s.
            simsobj.data[spc] *= ions_per_amp/(1e5 * dct['resistor'])

def _guess_fc_resistors(simsobj):
    """ Internal function; guess FC resistor value from raw and
        Cameca-corrected data. Returns a pandas Series with
        species labels as index.
    """
    sp = []
    bg = []
    ct = []
    for species, mt in simsobj.header['MassTable'].items():
        bfi = mt['b field index']
        tri = mt['trolley index']
        tr = simsobj.header['BFields'][bfi]['Trolleys'][tri]
        if tr['detector'] == 'FC':
            sp.append(species)
            dt = tr['detector label']
            if simsobj.header['polarity'] == '+':
                bg.append(simsobj.header['Detectors'][dt]['fc background setup positive'])
            else:
                bg.append(simsobj.header['Detectors'][dt]['fc background setup negative'])

    data = simsobj.data[sp].mean()
    corr = simsobj._data_corr[sp].mean()
    resistors = (data - bg) * ions_per_amp / (corr * 1e5)
    resistors = np.log10(resistors).round()
    resistors = 10**resistors
    resistors.index.name = 'Ohm'
    return resistors

class _JSONDateTimeEncoder(json.JSONEncoder):
    """ Converts datetime objects to json format. """
    def default(self, obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        else:
            return json.JSONEncoder.default(self, obj)


def export_header(simsobj, filename=""):
    """ Export header to a JSON text file.

        Usage: export_header(simsobj, filename="alt_filename.txt")

        The entire header will be pretty-printed to a text file,
        serialized as JSON in UTF-8 encoding. Uses sims_filename.im.txt
        by default.
    """
    if not filename:
        filename = simsobj.filename + '.txt'

    # io is for python2 compatibility
    with io.open(filename, mode='wt', encoding='utf-8') as fp:
        print(json.dumps(simsobj.header, sort_keys=True,
                         indent=2, ensure_ascii=False,
                         separators=(',', ': '), cls=_JSONDateTimeEncoder),
              file=fp)

def export_matlab(simsobj, filename="", prefix='m', **kwargs):
    """ Export data to MatLab file.

        Usage: export_matlab(simsobj, filename="alt_filename.mat")

        Saves data to filename.im.mat, or alt_filename.mat if supplied. All
        other keyword arguments are passed on to scipy.io.savemat().

        By default the MatLab file will be saved in MatLab 5 format (the default
        in MatLab 5 - 7.2), with long_field_names=True (compatible with
        MatLab 7.6+) and with do_compression=True.

        The data is stored as a dictionary with the labels as keywords and
        3D numpy arrays as values. Each label is taken from the nanoSIMS label,
        prefixed with 'prefix' (default 'm'), because MatLab doesn't allow
        variable names starting with a number.
    """
    try:
        import scipy.io
    except ImportError:
        msg = "Scipy not found on your system, MatLab export not available."
        warnings.warn(msg)
        return

    if not filename:
        filename = simsobj.filename
    if 'do_compression' not in kwargs.keys():
        kwargs['do_compression'] = True
    if 'long_field_names' not in kwargs.keys():
        kwargs['long_field_names'] = True

    export = {}
    if pd:
        for l in simsobj.data.labels:
            export[prefix + l] = np.asarray(simsobj.data[l])
    else:
        for n, l in enumerate(simsobj.header['label list']):
            export[prefix + l] = simsobj.data[n]

    scipy.io.savemat(filename, export, **kwargs)



""" Python module to read Cameca (nano)SIMS data files. """

from sims.sims import *

__version__ = '1.0.0'
__name__ = 'sims'
__author__ = 'Zan Peeters'
__url__ = 'https://github.com/zanpeeters/sims'
__license__ = 'BSD 3-Clause Clear'
__copyright__ = '(c) 2018 Zan Peeters'
__description__ = __doc__



#!/usr/bin/env python
""" Setuptools setup file for sims. """
from setuptools import setup, find_packages
import os

with open(os.path.join('sims', '__init__.py'), mode='rt', encoding='utf-8') as fh:
    script = []
    for l in fh.readlines():
        l = l.strip()
        # don't import anything, just get metadata
        if l.startswith('from') or l.startswith('import'):
            continue
        script.append(l)

exec('\n'.join(script))

with open('README.md', mode='rt', encoding='utf-8') as fh:
    __long_description__ = fh.read()

setup(
    name = __name__,
    version = __version__,
    description = __description__,
    long_description = __long_description__,
    url = __url__,
    author = __author__,
    author_email = 'me@example.com',
    license = __license__,
    classifiers = [
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering'
    ],
    keywords = 'sims nanosims mass-spectrometry Cameca file-format',

    install_requires = [
        'matplotlib',
        'pandas',
        'scipy'
    ],

    packages = find_packages(),
    package_data = {'sims': ['lut/*']},
    zip_safe = False
)
